#include <array>
#include <fstream>
#include <iostream>
#include <string>
#include <thread>
#include <vector>
#include "nmdl.h"
#include "nmdl_compiler.h"
#include "nmdl_image_converter.h"

namespace {

NMDL_RESULT Call(NMDL_COMPILER_RESULT result, const std::string &function_name) {
	switch(result) {
	case NMDL_COMPILER_RESULT_OK:
		return NMDL_RESULT_OK;
	case NMDL_COMPILER_RESULT_MEMORY_ALLOCATION_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_ALLOCATION_ERROR");
	case NMDL_COMPILER_RESULT_MODEL_LOADING_ERROR:
		throw std::runtime_error(function_name + ": MODEL_LOADING_ERROR");
	case NMDL_COMPILER_RESULT_INVALID_PARAMETER:
		throw std::runtime_error(function_name + ": INVALID_PARAMETER");
	case NMDL_COMPILER_RESULT_INVALID_MODEL:
		throw std::runtime_error(function_name + ": INVALID_MODEL");
	case NMDL_COMPILER_RESULT_UNSUPPORTED_OPERATION:
		throw std::runtime_error(function_name + ": UNSUPPORTED_OPERATION");
	default:
		throw std::runtime_error(function_name + ": UNKNOWN ERROR");
	}
}

NMDL_RESULT Call(NMDL_RESULT result, const std::string &function_name) {
	switch(result) {
	case NMDL_RESULT_OK:
		return NMDL_RESULT_OK;
	case NMDL_RESULT_INVALID_FUNC_PARAMETER:
		throw std::runtime_error(function_name + ": INVALID_FUNC_PARAMETER");
	case NMDL_RESULT_NO_LOAD_LIBRARY:
		throw std::runtime_error(function_name + ": NO_LOAD_LIBRARY");
	case NMDL_RESULT_NO_BOARD:
		throw std::runtime_error(function_name + ": NO_BOARD");
	case NMDL_RESULT_BOARD_RESET_ERROR:
		throw std::runtime_error(function_name + ": BOARD_RESET_ERROR");
	case NMDL_RESULT_INIT_CODE_LOADING_ERROR:
		throw std::runtime_error(function_name + ": INIT_CODE_LOADING_ERROR");
	case NMDL_RESULT_CORE_HANDLE_RETRIEVAL_ERROR:
		throw std::runtime_error(function_name + ": CORE_HANDLE_RETRIEVAL_ERROR");
	case NMDL_RESULT_FILE_LOADING_ERROR:
		throw std::runtime_error(function_name + ": FILE_LOADING_ERROR");
	case NMDL_RESULT_MEMORY_WRITE_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_WRITE_ERROR");
	case NMDL_RESULT_MEMORY_READ_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_READ_ERROR");
	case NMDL_RESULT_MEMORY_ALLOCATION_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_ALLOCATION_ERROR");
	case NMDL_RESULT_MODEL_LOADING_ERROR:
		throw std::runtime_error(function_name + ": MODEL_LOADING_ERROR");
	case NMDL_RESULT_INVALID_MODEL:
		throw std::runtime_error(function_name + ": INVALID_MODEL");
	case NMDL_RESULT_BOARD_SYNC_ERROR:
		throw std::runtime_error(function_name + ": BOARD_SYNC_ERROR");
	case NMDL_RESULT_BOARD_MEMORY_ALLOCATION_ERROR:
		throw std::runtime_error(function_name + ": BOARD_MEMORY_ALLOCATION_ERROR");
	case NMDL_RESULT_NN_CREATION_ERROR:
		throw std::runtime_error(function_name + ": NN_CREATION_ERROR");
	case NMDL_RESULT_NN_LOADING_ERROR:
		throw std::runtime_error(function_name + ": NN_LOADING_ERROR");
	case NMDL_RESULT_NN_INFO_RETRIEVAL_ERROR:
		throw std::runtime_error(function_name + ": NN_INFO_RETRIEVAL_ERROR");
	case NMDL_RESULT_MODEL_IS_TOO_BIG:
		throw std::runtime_error(function_name + ": MODEL_IS_TOO_BIG");
	case NMDL_RESULT_NOT_INITIALIZED:
		throw std::runtime_error(function_name + ": NOT_INITIALIZED");
	case NMDL_RESULT_INCOMPLETE:
		throw std::runtime_error(function_name + ": INCOMPLETE");
	case NMDL_RESULT_UNKNOWN_ERROR:
		throw std::runtime_error(function_name + ": UNKNOWN_ERROR");
	default:
		throw std::runtime_error(function_name + ": UNKNOWN ERROR");
	};
}

template <typename T>
std::vector<T> ReadFile(const std::string &filename) {
	std::ifstream ifs(filename, std::ios::binary | std::ios::ate);
	if(!ifs.is_open()) {
		throw std::runtime_error("Unable to open input file: " + filename);
	}
	auto fsize = static_cast<std::size_t>(ifs.tellg());
	ifs.seekg(0);
	std::vector<T> data(fsize / sizeof(T));
	ifs.read(reinterpret_cast<char*>(data.data()), data.size() * sizeof(T));
	return data;
}

void ShowNMDLVersion() {
	std::uint32_t major = 0;
	std::uint32_t minor = 0;
	std::uint32_t patch = 0;
	Call(NMDL_GetLibVersion(&major, &minor, &patch), "GetLibVersion");
	std::cout << "Lib version: " << major << "." << minor 
			<< "." << patch << std::endl;
}

void CheckBoard(std::uint32_t required_board_type) {
	std::uint32_t boards = 0;
	std::uint32_t board_number = -1;
	Call(NMDL_GetBoardCount(required_board_type, &boards), "GetBoardCount");
	std::cout << "Detected boards: " << boards << std::endl;
	if(!boards) {
		throw std::runtime_error("Board not found");
	}
}

std::vector<float> CompileModel(const std::string &model_filename, std::uint32_t board_type,
		bool is_multi_unit) {
	float *nm_model = nullptr;
	std::uint32_t nm_model_floats = 0u;
	auto model = ReadFile<char>(model_filename);
	Call(NMDL_COMPILER_CompileONNX(is_multi_unit, board_type, model.data(),
			static_cast<std::uint32_t>(model.size()), &nm_model, 
			&nm_model_floats), "CompileONNX");
	std::vector<float> result(nm_model, nm_model + nm_model_floats);
	NMDL_COMPILER_FreeModel(board_type, nm_model);
	return result;
}

NMDL_ModelInfo GetModelInformation(NMDL_HANDLE nmdl, std::uint32_t unit_num) {
	NMDL_ModelInfo model_info;
	Call(NMDL_GetModelInfo(nmdl, unit_num, &model_info), "GetModelInfo");
	std::cout << "Input tensor number: " << model_info.input_tensor_num << std::endl;
	for(std::size_t i = 0; i < model_info.input_tensor_num; ++i) {
		std::cout << "Input tensor " << i << ": " <<
			model_info.input_tensors[i].width << ", " <<
			model_info.input_tensors[i].height << ", " <<
			model_info.input_tensors[i].depth <<
			std::endl;
	}
	std::cout << "Output tensor number: " << model_info.output_tensor_num << std::endl;
	for(std::size_t i = 0; i < model_info.output_tensor_num; ++i) {
		std::cout << "Output tensor " << i << ": " <<
			model_info.output_tensors[i].width << ", " <<
			model_info.output_tensors[i].height << ", " <<
			model_info.output_tensors[i].depth <<
			std::endl;
	}
	return model_info;
}

std::vector<float> PrepareInput(const std::string &filename, std::uint32_t width, 
		std::uint32_t height, std::uint32_t board_type, 
		std::uint32_t color_format, const float rgb_divider[3], 
		const float rgb_adder[3]) {
	auto bmp_frame = ReadFile<char>(filename);
	std::vector<float> input(NMDL_IMAGE_CONVERTER_RequiredSize(
			width, height, color_format, board_type));
	if(NMDL_IMAGE_CONVERTER_Convert(bmp_frame.data(), input.data(), 
			static_cast<std::uint32_t>(bmp_frame.size()), width, height, 
			color_format, rgb_divider, rgb_adder, board_type)) {
		throw std::runtime_error("Image conversion error");
	}
	return input;
}

void WaitForOutput(NMDL_HANDLE nmdl, std::uint32_t unit_num, float *outputs[]) {
	std::uint32_t status = NMDL_PROCESS_FRAME_STATUS_INCOMPLETE;
	while(status == NMDL_PROCESS_FRAME_STATUS_INCOMPLETE) {
		NMDL_GetStatus(nmdl, unit_num, &status);
	};
	double fps;
	Call(NMDL_GetOutput(nmdl, unit_num, outputs, &fps), "GetOutput");
	std::cout << "First four result values:" << std::endl;
	for(std::size_t i = 0; i < 4; ++i) {
		std::cout << outputs[0][i] << std::endl;
	}
	std::cout << "FPS:" << fps << std::endl;
}

}

int main() {
	const std::size_t FRAMES = 10;
	const std::uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_SIMULATOR;
	const std::string MODEL_FILENAMES[NMDL_MAX_UNITS] = {
			"../nmdl_ref_data/squeezenet_imagenet/model.onnx",
			"../nmdl_ref_data/resnet_18_imagenet/model.onnx",
			"../nmdl_ref_data/resnet_50_imagenet/model.onnx",
			"../nmdl_ref_data/inception_v3_imagenet/model.onnx"
	};
	const std::string FRAME_FILENAMES[NMDL_MAX_UNITS] = {
			"../nmdl_ref_data/squeezenet_imagenet/frame.bmp",
			"../nmdl_ref_data/resnet_18_imagenet/frame.bmp",
			"../nmdl_ref_data/resnet_50_imagenet/frame.bmp",
			"../nmdl_ref_data/inception_v3_imagenet/frame.bmp"
	};
	const NMDL_IMAGE_CONVERTER_COLOR_FORMAT COLOR_FORMATS[NMDL_MAX_UNITS] = {
			NMDL_IMAGE_CONVERTER_COLOR_FORMAT_BGR,
			NMDL_IMAGE_CONVERTER_COLOR_FORMAT_BGR,
			NMDL_IMAGE_CONVERTER_COLOR_FORMAT_RGB,
			NMDL_IMAGE_CONVERTER_COLOR_FORMAT_RGB
	};
	const float NM_FRAME_RGB_DIVIDER[NMDL_MAX_UNITS][3] = {
			{1.0f, 1.0f, 1.0f},
			{1.0f, 1.0f, 1.0f},
			{1.0f, 1.0f, 1.0f},
			{127.5f, 127.5f, 127.5f}
	};
	const float NM_FRAME_RGB_ADDER[NMDL_MAX_UNITS][3] = {
			{0.0f, 0.0f, 0.0f},
			{0.0f, 0.0f, 0.0f},
			{-114.0f, -114.0f, -114.0f},
			{-1.0f, -1.0f, -1.0f}
	};

	NMDL_HANDLE nmdl = 0;

	try {
		std::cout << "Query library version..." << std::endl;
		ShowNMDLVersion();

		std::cout << "Board detection... " << std::endl;
		CheckBoard(BOARD_TYPE);
			
		std::cout << "NMDL initialization... " << std::endl;
		Call(NMDL_Create(&nmdl), "Create");
		
		std::array<std::vector<float>, NMDL_MAX_UNITS> models;
		for(std::size_t unit = 0; unit < NMDL_MAX_UNITS; ++unit) {
			std::cout << "Compile model " << unit << "... " << std::endl;
			models[unit] = CompileModel(MODEL_FILENAMES[unit], 
					NMDL_COMPILER_BOARD_TYPE_MC12705, false);
		}
		std::array<const float*, NMDL_MAX_UNITS> model_pointers = {
			models[0].data(), models[1].data(), models[2].data(), models[3].data()
		};
		std::array<std::uint32_t, NMDL_MAX_UNITS> model_floats = {
			static_cast<std::uint32_t>(models[0].size()),
			static_cast<std::uint32_t>(models[1].size()),
			static_cast<std::uint32_t>(models[2].size()),
			static_cast<std::uint32_t>(models[3].size())
		};
		Call(NMDL_Initialize(nmdl, BOARD_TYPE, 0, 0, model_pointers.data(), 
				model_floats.data()), "Initialize");

		std::array<std::vector<float>, NMDL_MAX_UNITS> inputs;
		std::array<std::vector<float>, NMDL_MAX_UNITS> outputs;
		for(std::uint32_t unit = 0; unit < NMDL_MAX_UNITS; ++unit) {
			std::cout << "Get model " << unit << " information... " << std::endl;
			auto model_info = GetModelInformation(nmdl, unit);

			std::cout << "Prepare input " << unit << " information... " << std::endl;
			inputs[unit] = PrepareInput(FRAME_FILENAMES[unit], model_info.input_tensors[0].width,
					model_info.input_tensors[0].height, NMDL_IMAGE_CONVERTER_BOARD_TYPE_MC12705,
					COLOR_FORMATS[unit], NM_FRAME_RGB_DIVIDER[unit], NM_FRAME_RGB_ADDER[unit]);

			std::cout << "Reserve outputs " << unit << " information... " << std::endl;
			outputs[unit].resize(model_info.output_tensors[0].width *
					model_info.output_tensors[0].height * model_info.output_tensors[0].depth);
		}

		std::array<std::thread, NMDL_MAX_UNITS>  threads;
		for(std::uint32_t unit = 0; unit < NMDL_MAX_UNITS; ++unit) {
			threads[unit] = std::thread([=]() {
				std::array<const float*, 1> input_pointer = {inputs[unit].data()};
				std::array<float*, 1> output_pointer = {const_cast<float*>(outputs[unit].data())};
				for(std::size_t frame = 0; frame < FRAMES; ++frame) {
					std::cout << "Thread " << unit << ": Process frame " << frame << "... " << std::endl;
					Call(NMDL_Process(nmdl, unit, input_pointer.data()), "Process");
					WaitForOutput(nmdl, unit, output_pointer.data());
				}
			});
		}
		for(std::size_t unit = 0; unit < NMDL_MAX_UNITS; ++unit) {
			threads[unit].join();
		}
	}
	catch (std::exception& e) {
		std::cerr << e.what() << std::endl;
	}

	NMDL_Release(nmdl);
	NMDL_Destroy(nmdl);

	return 0;
}
