#include <array>
#include <chrono>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include "nmdl.h"

namespace {

NMDL_RESULT Call(NMDL_RESULT result, const std::string &function_name) {
	switch(result) {
	case NMDL_RESULT_OK:
		return NMDL_RESULT_OK;
	case NMDL_RESULT_INVALID_FUNC_PARAMETER:
		throw std::runtime_error(function_name + ": INVALID_FUNC_PARAMETER");
	case NMDL_RESULT_NO_LOAD_LIBRARY:
		throw std::runtime_error(function_name + ": NO_LOAD_LIBRARY");
	case NMDL_RESULT_NO_BOARD:
		throw std::runtime_error(function_name + ": NO_BOARD");
	case NMDL_RESULT_BOARD_RESET_ERROR:
		throw std::runtime_error(function_name + ": BOARD_RESET_ERROR");
	case NMDL_RESULT_INIT_CODE_LOADING_ERROR:
		throw std::runtime_error(function_name + ": INIT_CODE_LOADING_ERROR");
	case NMDL_RESULT_CORE_HANDLE_RETRIEVAL_ERROR:
		throw std::runtime_error(function_name + ": CORE_HANDLE_RETRIEVAL_ERROR");
	case NMDL_RESULT_FILE_LOADING_ERROR:
		throw std::runtime_error(function_name + ": FILE_LOADING_ERROR");
	case NMDL_RESULT_MEMORY_WRITE_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_WRITE_ERROR");
	case NMDL_RESULT_MEMORY_READ_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_READ_ERROR");
	case NMDL_RESULT_MEMORY_ALLOCATION_ERROR:
		throw std::runtime_error(function_name + ": MEMORY_ALLOCATION_ERROR");
	case NMDL_RESULT_MODEL_LOADING_ERROR:
		throw std::runtime_error(function_name + ": MODEL_LOADING_ERROR");
	case NMDL_RESULT_INVALID_MODEL:
		throw std::runtime_error(function_name + ": INVALID_MODEL");
	case NMDL_RESULT_BOARD_SYNC_ERROR:
		throw std::runtime_error(function_name + ": BOARD_SYNC_ERROR");
	case NMDL_RESULT_BOARD_MEMORY_ALLOCATION_ERROR:
		throw std::runtime_error(function_name + ": BOARD_MEMORY_ALLOCATION_ERROR");
	case NMDL_RESULT_NN_CREATION_ERROR:
		throw std::runtime_error(function_name + ": NN_CREATION_ERROR");
	case NMDL_RESULT_NN_LOADING_ERROR:
		throw std::runtime_error(function_name + ": NN_LOADING_ERROR");
	case NMDL_RESULT_NN_INFO_RETRIEVAL_ERROR:
		throw std::runtime_error(function_name + ": NN_INFO_RETRIEVAL_ERROR");
	case NMDL_RESULT_MODEL_IS_TOO_BIG:
		throw std::runtime_error(function_name + ": MODEL_IS_TOO_BIG");
	case NMDL_RESULT_NOT_INITIALIZED:
		throw std::runtime_error(function_name + ": NOT_INITIALIZED");
	case NMDL_RESULT_INCOMPLETE:
		throw std::runtime_error(function_name + ": INCOMPLETE");
	case NMDL_RESULT_UNKNOWN_ERROR:
		throw std::runtime_error(function_name + ": UNKNOWN_ERROR");
	default:
		throw std::runtime_error(function_name + ": UNKNOWN ERROR");
	};
}

template <typename T>
std::vector<T> ReadFile(const std::string &filename) {
	std::ifstream ifs(filename, std::ios::binary | std::ios::ate);
	if(!ifs.is_open()) {
		throw std::runtime_error("Unable to open input file: " + filename);
	}
	auto fsize = static_cast<std::size_t>(ifs.tellg());
	ifs.seekg(0);
	std::vector<T> data(fsize / sizeof(T));
	ifs.read(reinterpret_cast<char*>(data.data()), data.size() * sizeof(T));
	return data;
}

void ShowNMDLVersion() {
	std::uint32_t major = 0;
	std::uint32_t minor = 0;
	std::uint32_t patch = 0;
	Call(NMDL_GetLibVersion(&major, &minor, &patch), "GetLibVersion");
	std::cout << "Lib version: " << major << "." << minor 
			<< "." << patch << std::endl;
}

auto GetBoardCount(std::uint32_t required_board_type) {
	std::uint32_t boards = 0;
	Call(NMDL_GetBoardCount(required_board_type, &boards), "GetBoardCount");
	std::cout << "Detected boards: " << boards << std::endl;
	if(!boards) {
		throw std::runtime_error("Board not found");
	}
	return boards;
}

NMDL_ModelInfo GetModelInformation(NMDL_HANDLE nmdl, std::uint32_t unit_num) {
	NMDL_ModelInfo model_info;
	Call(NMDL_GetModelInfo(nmdl, unit_num, &model_info), "GetModelInfo");
	std::cout << "Mode: " << (model_info.is_multi_unit ? "multi" : "single") << " unit" << std::endl;
	std::cout << "Input tensor number: " << model_info.input_tensor_num << std::endl;
	for(std::size_t i = 0; i < model_info.input_tensor_num; ++i) {
		std::cout << "Input tensor " << i << ": " <<
			model_info.input_tensors[i].width << ", " <<
			model_info.input_tensors[i].height << ", " <<
			model_info.input_tensors[i].depth <<
			std::endl;
	}
	std::cout << "Output tensor number: " << model_info.output_tensor_num << std::endl;
	for(std::size_t i = 0; i < model_info.output_tensor_num; ++i) {
		std::cout << "Output tensor " << i << ": " <<
			model_info.output_tensors[i].width << ", " <<
			model_info.output_tensors[i].height << ", " <<
			model_info.output_tensors[i].depth <<
			std::endl;
	}
	return model_info;
}

double WaitForOutput(NMDL_HANDLE nmdl, std::uint32_t unit_num, float *outputs[]) {
	std::uint32_t status = NMDL_PROCESS_FRAME_STATUS_INCOMPLETE;
	while(status == NMDL_PROCESS_FRAME_STATUS_INCOMPLETE) {
		NMDL_GetStatus(nmdl, unit_num, &status);
	};
	double fps;

	Call(NMDL_GetOutput(nmdl, unit_num, outputs, &fps), "GetOutput");
	//std::cout << "First four result values:" << std::endl;
	//for(std::size_t i = 0; i < 4; ++i) {
	//	std::cout << outputs[0][i] << std::endl;
	//}
	//std::cout << "FPS:" << fps << std::endl;

	return fps;
}

#define CNT_2_BOARD(CNT, UNITS, BOARDS) ((CNT) / (UNITS)) % (BOARDS) 
#define CNT_2_UNIT(CNT, UNITS) (CNT) % (UNITS) 

}

int main() {
	const std::uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_SIMULATOR;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_MC12705;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_NMCARD;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_NMMEZZO;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_NMQUAD;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_MC12101;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_NMSTICK;
	//const uint32_t BOARD_TYPE = NMDL_BOARD_TYPE_NMEMBED;
	const std::string MODEL_FILENAME =
		"../nmdl_ref_data/squeezenet_imagenet/model_mu.nm8";
	const std::string FRAME_FILENAME =
		"../nmdl_ref_data/squeezenet_imagenet/frame08";
	const std::size_t FRAMES = 100;

	std::vector<NMDL_HANDLE> nmdl;

	try {
		std::cout << "Query library version..." << std::endl;
		ShowNMDLVersion();

		std::cout << "Boards detection... " << std::endl;
		nmdl.resize(GetBoardCount(BOARD_TYPE));
			
		auto model = ReadFile<float>(MODEL_FILENAME);
		std::array<const float*, NMDL_MAX_UNITS> models = {
				model.data(), 
				model.data(), 
				model.data(), 
				model.data()
		};
		std::array<std::uint32_t, NMDL_MAX_UNITS> model_floats = {
				static_cast<std::uint32_t>(model.size()),
				static_cast<std::uint32_t>(model.size()),
				static_cast<std::uint32_t>(model.size()),
				static_cast<std::uint32_t>(model.size())
		};

		for(std::uint32_t cur_board = 0; cur_board < nmdl.size(); ++cur_board) {
			std::cout << "Board " << cur_board << "NMDL initialization... " << std::endl;
			Call(NMDL_Create(&nmdl[cur_board]), "Create");
			Call(NMDL_Initialize(nmdl[cur_board], BOARD_TYPE, cur_board, 0, models.data(),
				model_floats.data()), "Initialize");
		}

		std::cout << "Get model information... " << std::endl;
		auto model_info = GetModelInformation(nmdl[0], 0);

		std::cout << "Prepare inputs... " << std::endl;
		auto input = ReadFile<float>(FRAME_FILENAME);
		std::array<const float*, 1> inputs = {input.data()};

		std::cout << "Reserve outputs... " << std::endl;
		std::vector<std::vector<float>> output_tensors(model_info.output_tensor_num);
		std::vector<float*> outputs(model_info.output_tensor_num);
		for(std::size_t i = 0; i < model_info.output_tensor_num; ++i) {
			output_tensors[i].resize(static_cast<std::size_t>(
					model_info.output_tensors[i].width) *
					model_info.output_tensors[i].height * 
					model_info.output_tensors[i].depth);
			outputs[i] = output_tensors[i].data();
		}

		std::cout << "Process inputs... " << std::endl;
		std::uint32_t proc_batches = model_info.is_multi_unit ? 1 : 4;
		std::uint32_t total_batches = nmdl.size() * proc_batches > FRAMES ?
				FRAMES : nmdl.size() * proc_batches;
		std::uint32_t cnt_in = 0;
		std::uint32_t cnt_out = 0;
		double fps = 0.0;
		auto start_time = std::chrono::system_clock::now();
		for(auto i = 0u; i < total_batches; ++i, cnt_in++) {
			Call(NMDL_Process(
					nmdl[CNT_2_BOARD(cnt_in, proc_batches, nmdl.size())],
					CNT_2_UNIT(cnt_in, proc_batches), 
					inputs.data()), "ProcessFrame");
		}
		for(auto i = total_batches; i < FRAMES; ++i, cnt_in++, cnt_out++) {
			fps += WaitForOutput(
					nmdl[CNT_2_BOARD(cnt_out, proc_batches, nmdl.size())],
					CNT_2_UNIT(cnt_out, proc_batches), outputs.data());
			Call(NMDL_Process(
					nmdl[CNT_2_BOARD(cnt_in, proc_batches, nmdl.size())],
					CNT_2_UNIT(cnt_in, proc_batches), 
					inputs.data()), "ProcessFrame");
		}
		for(auto i = 0u; i < total_batches; ++i, cnt_out++) {
			fps += WaitForOutput(
					nmdl[CNT_2_BOARD(cnt_out, proc_batches, nmdl.size())],
					CNT_2_UNIT(cnt_out, proc_batches), outputs.data());
		}
		auto delta_time = std::chrono::duration_cast<std::chrono::
				milliseconds>(std::chrono::system_clock::now() - start_time).count();	
		std::cout << "Elapset time (ms): " << delta_time << std::endl;
		std::cout << "FPS: " << (1000.0 * FRAMES / delta_time) << std::endl;
		std::cout << "FPS on board: " << ((fps / FRAMES) * proc_batches) << std::endl;
	}
	catch (std::exception& e) {
		std::cerr << e.what() << std::endl;
	}

	for(auto cur : nmdl) {
		NMDL_Release(cur);
		NMDL_Destroy(cur);
	}

	return 0;
}
